#include <zlib.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "gzip.h"
#include "buff.h"

unsigned char gzipHead[] =
{
  0x1f,         // Magic number
  0x8b,         // Magic number
  0x08,         // Compression method
  0x00,         // Flags
  0x00,         // Modification time
  0x00,         // Modification time
  0x00,         // Modification time
  0x00,         // Modification time
  0x00,         // Extra flags
  0xFF          // Operating system
};

buff_t *
gzip_Compress (data, leng)
  unsigned const char * data;
  int leng;
{
  int status = 0;
  buff_t * buff;
  
  unsigned char * destBuf;
  unsigned long   destLen;
  unsigned long   dataSum;
  unsigned long   dataLen;
  unsigned char   sumBuf[4];
  unsigned char   lenBuf[4];
  
  assert(data != NULL);
  
  destLen = 12 + (leng * 1.1);
  
  destBuf = (unsigned char *)malloc(destLen);
  assert(destBuf != NULL);
  bzero(destBuf, destLen);
  
  if (Z_OK != (status = compress2(destBuf, &destLen, data, leng, 9))) {
    fprintf(stderr, "gzip_Compress.. failed to compress, %d\n", status);
    free(destBuf);
    return NULL;
  }
  
  buff = buff_Create();
//buff_Init(buff, leng * 2);
  buff_Init(buff, 1024 * 1024);
  buff_Append(buff, gzipHead, sizeof(gzipHead));
  buff_Append(buff, destBuf + 2, destLen - 2);
  
  dataSum = gzip_Checksum(data, leng);
  
  dataSum = htonl(dataSum);
  dataLen = htonl(leng);

  memcpy(sumBuf, &dataSum, 4);
  memcpy(lenBuf, &dataLen, 4);
  
  buff_Append(buff, (unsigned char *)&sumBuf[3], 1);
  buff_Append(buff, (unsigned char *)&sumBuf[2], 1);
  buff_Append(buff, (unsigned char *)&sumBuf[1], 1);
  buff_Append(buff, (unsigned char *)&sumBuf[0], 1);

  buff_Append(buff, (unsigned char *)&lenBuf[3], 1);
  buff_Append(buff, (unsigned char *)&lenBuf[2], 1);
  buff_Append(buff, (unsigned char *)&lenBuf[1], 1);
  buff_Append(buff, (unsigned char *)&lenBuf[0], 1);

  free(destBuf);
  
  return buff;
}

unsigned long
gzip_Checksum (data, leng)
  unsigned const char * data;
  int leng;
{
  unsigned long crc;
  
  crc = crc32(0L, Z_NULL, 0);
  crc = crc32(crc, data, leng);
  
  return crc;
}
