#ifndef __HASH_H
#define __HASH_H

#include <pthread.h>

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** The Various Structs
#endif

typedef struct                // the key structure - stored in a bucket
{                             //
  unsigned char * name;       // the name of the key
  unsigned char * data;       // the data stored by the key
  int nleng;                  // length (in bytes) of name
  int dleng;                  // length (in bytes) of data
} hash_key_t;

typedef struct                // the bucket structure - stored in a hash
{                             //
  pthread_mutex_t mutex;      // for thread safeness
  int leng;                   // number of keys in bucket
  int size;                   // malloc'ed size of bucket
  hash_key_t ** list;         // the list of keys
} hash_bucket_t;

typedef struct                // the hash structure itself
{                             //
  pthread_mutex_t mutex;      // for thread safeness
  int leng;                   // number of buckets in hash
  unsigned long size;         // malloc'ed size of bucket
  hash_bucket_t ** list;      // the list of buckets
} hash_t;

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** hash_t functions
#endif

hash_t * hash_Create ();
void hash_Init (hash_t *);
inline unsigned long hash_Hash (hash_t *, unsigned const char *, unsigned long);
hash_key_t * hash_Put (hash_t *, unsigned const char * name, unsigned long nleng, unsigned const char * data, unsigned long dleng);
hash_key_t * hash_Get (hash_t *, unsigned const char * name, unsigned long nleng);
void hash_Remove (hash_t *, unsigned const char * name, unsigned long nleng);
void hash_Destroy (hash_t *);

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** bucket_t functions
#endif

hash_bucket_t * hash_bucket_Create ();
void hash_bucket_Init (hash_bucket_t *);
hash_key_t * hash_bucket_AppendValue (hash_bucket_t *, unsigned const char * name, unsigned long nleng, unsigned const char * data, unsigned long dleng);
hash_key_t * hash_bucket_AppendKey (hash_bucket_t *, hash_key_t *);
hash_key_t * hash_bucket_GetByValue (hash_bucket_t *, unsigned const char * name, unsigned long nleng);
void hash_bucket_RemoveByValue (hash_bucket_t *, unsigned const char * name, unsigned long nleng);
void hash_bucket_Destroy (hash_bucket_t *);

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** key_t functions
#endif

hash_key_t * hash_key_Create ();
void hash_key_Init (hash_key_t *, unsigned const char * name, unsigned long nleng, unsigned const char * data, unsigned long dleng);
int hash_key_CompareToValue (hash_key_t *, unsigned const char * name, unsigned long nleng);
int hash_key_Compare (hash_key_t *, hash_key_t *);
void hash_key_SetValue (hash_key_t *, unsigned const char * data, unsigned long dleng);
void hash_key_Destroy (hash_key_t *);

#endif /* __HASH_H */
