#include <stdlib.h>       // malloc()
#include <assert.h>       // assert()
#include <strings.h>      // bzero()
#include <stdio.h>        // fopen()
#include <sys/types.h>    // fstat()
#include <sys/stat.h>     // fstat();

// #include <id3/tag.h>   // id3lib

#include "id3.h"
#include "daap.h"

static const char * id3_v2_genres[] =
{
  "Blues",
  "Classic Rock",
  "Country",
  "Dance",
  "Disco",
  "Funk",
  "Grunge",
  "Hip-Hop",
  "Jazz",
  "Metal",
  "New Age",
  "Oldies",
  "Other",
  "Pop",
  "R&B",
  "Rap",
  "Reggae",
  "Rock",
  "Techno",
  "Industrial",
  "Alternative",
  "Ska",
  "Death Metal",
  "Pranks",
  "Soundtrack",
  "Euro-Techno",
  "Ambient",
  "Trip-Hop",
  "Vocal",
  "Jazz+Funk",
  "Fusion",
  "Trance",
  "Classical",
  "Instrumental",
  "Acid",
  "House",
  "Game",
  "Sound Clip",
  "Gospel",
  "Noise",
  "AlternRock",
  "Bass",
  "Soul",
  "Punk",
  "Space",
  "Meditative",
  "Instrumental Pop",
  "Instrumental Rock",
  "Ethnic",
  "Gothic",
  "Darkwave",
  "Techno-Industrial",
  "Electronic",
  "Pop-Folk",
  "Eurodance",
  "Dream",
  "Southern Rock",
  "Comedy",
  "Cult",
  "Gangsta",
  "Top 40",
  "Christian Rap",
  "Pop/Funk",
  "Jungle",
  "Native American",
  "Cabaret",
  "New Wave",
  "Psychadelic",
  "Rave",
  "Showtunes",
  "Trailer",
  "Lo-Fi",
  "Tribal",
  "Acid Punk",
  "Acid Jazz",
  "Polka",
  "Retro",
  "Musical",
  "Rock & Roll",
  "Hard Rock",
  "Folk",
  "Folk-Rock",
  "National Folk",
  "Swing",
  "Fast Fusion",
  "Bebob",
  "Latin",
  "Revival",
  "Celtic",
  "Bluegrass",
  "Avantgarde",
  "Gothic Rock",
  "Progressive Rock",
  "Psychedelic Rock",
  "Symphonic Rock",
  "Slow Rock",
  "Big Band",
  "Chorus",
  "Easy Listening",
  "Acoustic",
  "Humour",
  "Speech",
  "Chanson",
  "Opera",
  "Chamber Music",
  "Sonata",
  "Symphony",
  "Booty Bass",
  "Primus",
  "Porn Groove",
  "Satire",
  "Slow Jam",
  "Club",
  "Tango",
  "Samba",
  "Folklore",
  "Ballad",
  "Power Ballad",
  "Rhythmic Soul",
  "Freestyle",
  "Duet",
  "Punk Rock",
  "Drum Solo",
  "A capella",
  "Euro-House",
  "Dance Hall"
};

id3_data *
id3_Create ()
{
  id3_data * data;
  
  data = (id3_data *)malloc(sizeof(id3_data));
  assert(data != NULL);
  
  return data;
}

void
id3_Init (data)
  id3_data * data;
{
  assert(data != NULL);
  
  bzero(&data->fileident, 4);
  
  data->version_major = 0;
  data->version_minor = 0;
  data->size          = 0;
  
  data->exth          = NULL;
  
  data->frame_cnt     = 0;
  data->frame_max     = 0;
  data->frames        = NULL;
}

int
id3_Load_v22 (song, file)
  daap_song * song;
  FILE * file;
{
  return -1;
}

int
id3_Load_v23 (song, file)
  daap_song * song;
  FILE * file;
{
  assert(song != NULL);
  assert(file != NULL);
  
  fprintf(stderr, "id3_Load_v23.. unsupported version\n");
  fclose(file);
  
  return -1;
}

int
id3_Load_v24 (song, file)
  daap_song * song;
  FILE * file;
{
  assert(song != NULL);
  assert(file != NULL);
  
  fprintf(stderr, "id3_Load_v24.. unsupported version\n");
  fclose(file);
  
  return -1;
}

int
id3_Load (song, path)
  daap_song * song;
  char      * path;
{
  int leng = 0;
  int size = 0;
  FILE * file = NULL;
  char head[10];
  struct stat filestat;
  int fleng;
  int dleng;
  int rleng;
  
  int i_pos;
  int i_open;
  int i_close;
  int i_beg;
  int i_mid;
  int i_end;
  
  int pos;
  unsigned char fhead[6];      // frame header
  unsigned char fdata[255];    // frame data
  
  assert(song != NULL);
  assert(path != NULL);
  
  bzero(&head, 10);

  bzero(&filestat, sizeof(struct stat));

  // try to open the file. return -1 if we fail in this endeavor.
  if (NULL == (file = fopen(path, "r"))) {
    fprintf(stderr, "id3_Load.. could not open file, '%s'\n", path);
    return -1;
  }
  
  // fstat() returns all sorts of nifty information about the file in question.
  // particularly, we're interested in the file size (in bytes), and the last
  // modification time (in milliseconds) since Jan 1, 1970.
  fstat(fileno(file), &filestat);
  assz_Init(&song->assz, filestat.st_size);
  asdm_Init(&song->asdm, (daap_date)filestat.st_mtime);

  // the ID3 header, which I expect to be at the start of the file, is
  // exactly 10 bytes long. try to read it, and return -1 if we fail.
  if (10 != (leng = fread(&head, 1, 10, file))) {
    fprintf(stderr, "id3_Load.. could not read header, %d\n", leng);
    fclose(file);
    return -1;
  }

  // the first three bytes of an ID3 header serve to identify it as such.
  if (head[0] != 'I' || head[1] != 'D' || head[2] != '3') {
    fprintf(stderr, "id3_Load.. could not find ID3 marker, 0x%02hhx%02hhx%02hhx\n",
            head[0], head[1], head[2]);
    fclose(file);
    return -1;
  }

  // if (head[3] != 2) {
  //   fprintf(stderr, "id3_Load.. unsupported version, '%d.%d'\n", (int)head[3], (int)head[4]);
  //   fclose(file);
  //   return -1;
  // }
  
  switch ((int)head[3])
  {
    case 2:
      id3_Load_v22(song, file);
      break;
    case 3:
      id3_Load_v23(song, file);
      return -1;
      break;
    case 4:
      id3_Load_v24(song, file);
      return -1;
      break;
    default:
      fprintf(stderr, "id3_Load.. unsupported version, '%d.%d'\n", (int)head[3], (int)head[4]);
      fclose(file);
      return -1;
  }
  
  size  = ((int)head[6] << 24) >> 3;
  size += ((int)head[7] << 16) >> 2;
  size += ((int)head[8] <<  8) >> 1;
  size +=  (int)head[9];
  size += 10;

  pos = 10;
  
  while (1)
  {
    if (pos >= size)
      break;
    
    fleng = 0;          // frame length
    dleng = 0;
    rleng = fread(&fhead, 1, 6, file);
    
    if (rleng != 6) {
      fprintf(stderr, "id3_Load.. %s: failed to read six byte frame header, %d\n", path, rleng);
      break;
    }

    fleng  = fhead[3] << 16;
    fleng += fhead[4] <<  8;
    fleng += fhead[5];

    fhead[3] = 0;

    bzero(&fdata, 255);

    if (fhead[0] == 0)
      break;

    // fprintf(stderr, "id3_Load.. %s: new frame, NAME=%c%c%c, LENG=%d @ %d\n",
    //         path, fhead[0], fhead[1], fhead[2], fleng, pos);

    // if the three-character ID3 identifier is not soley comprised of capital
    // letters, A..Z, then we have probably encountered a problem, and should
    // exit now.
    if (fhead[0]<48 || fhead[0]>90 || fhead[1]<48 || fhead[1]>90 || fhead[2]<48 || fhead[2]>90) {
      fprintf(stderr, "id3_Load.. %s: non-alpha-numeric identifier ... probably an error\n", path);
      fprintf(stderr, "id3_Load.. %s: 0=%d, 1=%d, 2=%d\n", path, fhead[0], fhead[1], fhead[2]);
      break;
    }

    // if the stated leng of this ID3 frame, plus the current start-position of 
    // the ID3 frame, plus 6 for the size of the frame header, is greater than
    // the originally stated TOTAL size of all of the ID3 data, then we have
    // probably stumbled across a problem, and should exit now.
    if (fleng + pos + 6 > size) {
      fprintf(stderr, "id3_Load.. %s: length too long ... probably an error, %d\n", path, fleng);
      break;
    }
    
    if (0 == strncmp((char *)&fhead, "TT2", 3)) {
      dleng = fread(&fdata, 1, (fleng < 255 ? fleng : 255), file);
      minm_Init(&song->minm, (char *)&fdata + 1);
//    fprintf(stderr, "id3_Load.. SONG TITLE: '%s'\n", (char *)&fdata + 1);
    }
    else if (0 == strncmp((char *)&fhead, "TP1", 3)) {
      dleng = fread(&fdata, 1, (fleng < 255 ? fleng : 255), file);
      asar_Init(&song->asar, (char *)&fdata + 1);
//    fprintf(stderr, "id3_Load.. SONG ARTIST: '%s'\n", (char *)&fdata + 1);
    }
    else if (0 == strncmp((char *)&fhead, "TAL", 3)) {
      dleng = fread(&fdata, 1, (fleng < 255 ? fleng : 255), file);
      asal_Init(&song->asal, (char *)&fdata + 1);
//    fprintf(stderr, "id3_Load.. SONG ALBUM: '%s'\n", (char *)&fdata + 1);
    }
    else if (0 == strncmp((char *)&fhead, "TCO", 3)) {
      dleng = fread(&fdata, 1, (fleng < 255 ? fleng : 255), file);

      i_pos   =  0;
      i_open  = -1;
      i_close = -1;

      for (i_pos = 0; i_pos < dleng; i_pos++) {
        if (fdata[i_pos] == '(') {
          i_open = i_pos;
          break;
        }
      }

      // if the genre does not start with an open-paren (ie, '('), then just send the
      // literal value (less any zero-padding on either end) as the genre.
      if (i_open < 0) {
        // fprintf(stderr, "id3_Load.. %s: could not find open paren for genre\n", path);
        asgn_Init(&song->asgn, (char *)&fdata + 1);
        continue;
      }
      
      for (i_pos = i_open + 1; i_pos < dleng; i_pos++) {
        if (fdata[i_pos] == ')') {
          i_close = i_pos;
          break;
        }
      }

      if (i_close < 0) {
        fprintf(stderr, "id3_Load.. %s: could not find close paren for genre\n", path);
        continue;
      }
      
      fdata[i_close] = 0;

      // fprintf(stderr, "id3_Load.. %s: found genre number, %d\n", path, atoi((char *)&fdata+i_open+1));
      
//    asgn_Init(&song->asgn, (char *)&fdata + 1);
      asgn_Init(&song->asgn, (char *)id3_v2_genres[atoi((char *)&fdata+i_open+1)]);
//    fprintf(stderr, "id3_Load.. SONG GENRE: '%s'\n", (char *)&fdata + 1);
    }
    else if (0 == strncmp((char *)&fhead, "TYE", 3)) {
      dleng = fread(&fdata, 1, (fleng < 255 ? fleng : 255), file);
      asyr_Init(&song->asyr, (daap_short)atoi((char *)&fdata + 1));
//    fprintf(stderr, "id3_Load.. SONG ARTIST: '%s'\n", (char *)&fdata + 1);
    }
    else if (0 == strncmp((char *)&fhead, "TRK", 3)) {
      dleng = fread(&fdata, 1, (fleng < 255 ? fleng : 255), file);
      
      // given a string like: 5/13, with a possible binary zero on
      // either end, find the start of the first number, the separating
      // slash, and the end of the last number, then parse appropriately.
      i_pos =  0;
      i_beg = -1;
      i_mid = -1;
      i_end = -1;

      for (i_pos = 0; i_pos < dleng && i_beg == -1; i_pos++) {
        if (fdata[i_pos] >= 48 && fdata[i_pos] <= 57)
          i_beg = i_pos;
      }
      
      for (i_pos = i_beg + 1; i_pos < dleng && i_mid == -1; i_pos++) {
        if (fdata[i_pos] == '/')
          i_mid = i_pos;
      }

      // if there is a '/' (thus a track number and count), then set it
      // to zero so that atoi() can read the first number properly
      if (i_mid > 0)
        fdata[i_mid] = 0;

      // read in the track number
      astn_Init(&song->astn, (daap_short)atoi((char *)&fdata + i_beg));

      if (i_mid > 0) {
        for (i_pos = i_mid + 1; i_mid < dleng; i_pos++) {
          if (fdata[i_pos] >= 48 && fdata[i_pos] <= 57)
            i_end = i_pos;
          else
            break;
        }

        if (i_end > i_mid)
          astc_Init(&song->astc, (daap_short)atoi((char *)&fdata + i_mid + 1));
      }
    }
    else if (0 == strncmp((char *)&fhead, "TPA", 3)) {
      dleng = fread(&fdata, 1, (fleng < 255 ? fleng : 255), file);
      
      // given a string like: 5/13, with a possible binary zero on
      // either end, find the start of the first number, the separating
      // slash, and the end of the last number, then parse appropriately.
      i_pos =  0;
      i_beg = -1;
      i_mid = -1;
      i_end = -1;
      
      for (i_pos = 0; i_pos < dleng && i_beg == -1; i_pos++) {
        if (fdata[i_pos] >= 48 && fdata[i_pos] <= 57)
          i_beg = i_pos;
      }
      
      for (i_pos = i_beg + 1; i_pos < dleng && i_mid == -1; i_pos++) {
        if (fdata[i_pos] == '/')
          i_mid = i_pos;
      }
      
      // if there is a '/' (thus a track number and count), then set it
      // to zero so that atoi() can read the first number properly
      if (i_mid > 0)
        fdata[i_mid] = 0;
      
      // read in the track number
      asdn_Init(&song->asdn, (daap_short)atoi((char *)&fdata + i_beg));
      
      if (i_mid > 0) {
        for (i_pos = i_mid + 1; i_mid < dleng; i_pos++) {
          if (fdata[i_pos] >= 48 && fdata[i_pos] <= 57)
            i_end = i_pos;
          else
            break;
        }
        
        if (i_end > i_mid)
          asdc_Init(&song->asdc, (daap_short)atoi((char *)&fdata + i_mid + 1));
      }
    }
    
    if (0 != fseek(file, fleng - dleng, SEEK_CUR))
      break;

    pos += 6 + fleng;
  }
  
  

  fclose(file);
  
  // return pos;
  return size;
}

void
id3_Destroy (data)
  id3_data * data;
{
  assert(data != NULL);
  
}
