#ifndef __ID3_H
#define __ID3_H

#include <stdio.h>

#include "daap.h"

// these are the four possible flags that can be present
// in the ten-byte header of an ID3 structure. these bytes
// represent bits seven through four of a byte, and the
// remaining four bits most all be set to zero.
//
#define ID3_FLAG_UNSYNC   64          // bit 7 - unsynchronization
#define ID3_FLAG_EXTHEAD  32          // bit 6 - extended header
#define ID3_FLAG_EXPERM   16          // bit 5 - experimental indictator
#define ID3_FLAG_FOOTER    8          // bit 4 - footer present

typedef struct
{
  char frameid[4];                // four-character frame identifier
  unsigned int size;              // thirty-two bit synch safe frame size
  int flag_status;                // one byte - flags for status messages
  int flag_extra;                 // one byte - flags for format information
  unsigned char * data;           // actual frame data
} id3_frame;

typedef struct
{
  unsigned int size;              // thirty-two bit synch safe extended header size
  int flag_bytes;
  unsigned char * data;
} id3_exth;

typedef struct
{
  // required header data
  char fileident[3];              // file identifier (always "ID3")
  int version_major;              // one byte, major version number
  int version_minor;              // one byte, minor version number
  unsigned int size;              // thirty-two bit synch safe int

  id3_exth * exth;                // optional extended header
  
  // id3 frame list
  int frame_cnt;
  int frame_max;
  id3_frame ** frames;
} id3_data;

id3_data * id3_Create ();
void id3_Init (id3_data *);
int id3_Load (daap_song *, char * path);
int id3_Load_v22 (daap_song *, FILE *);
int id3_Load_v23 (daap_song *, FILE *);
int id3_Load_v24 (daap_song *, FILE *);
void id3_Destroy (id3_data *);

#endif /* __ID3_H */
