#include <stdio.h>            // fprintf(), fopen()
#include <assert.h>           // assert()
#include <signal.h>           // for sig stuff
#include <unistd.h>           // getopt(), etc
#include <stdlib.h>           // exit(), etc

#include "daap.h"
#include "http.h"
#include "mdns.h"
#include "hash.h"

static int __hault = 0;

daap_database kDefaultDatabase;
daap_adbs kDefaultAdbs;
daap_apso kDefaultApso;

static void
main_HandleSigPipe (sig)
  int sig;
{
  // fprintf(stderr, "main_HandleSigPipe\n");
  
  assert(sig == SIGPIPE);
}

static void
main_HandleSigUsr1 (sig)
  int sig;
{
  fprintf(stderr, "main_HandleSigUsr1()\n");

  assert(sig == SIGUSR1);
  
  daap_database_Update(&kDefaultDatabase, "./files");
}

static void
main_HandleSigUsr2 (sig)
  int sig;
{
  fprintf(stderr, "main_HandleSigUsr2()\n");

  // TODO: maybe this could act like a toggle switch, turning on
  // and off the ability for clients to execute further queries.
  // just an idea....
  
  assert(sig == SIGUSR2);
}

static void
main_HandleSigHup (sig)
{
  fprintf(stderr, "main_HandleSigHup()\n");
  
  assert(sig = SIGHUP);
}

static void
main_HandleSigInt (sig)
{
  fprintf(stderr, "main_HandleSigInt()\n");
  
  assert(sig == SIGINT);
  __hault = 1;
}

static void
main_HandleSigQuit (sig)
  int sig;
{
  fprintf(stderr, "main_HandleSigQuit()\n");
  
  assert(sig == SIGQUIT);
  exit(0);
}

/**
static void
main_ParseArguments (argc, argv)
  int argc;
  char ** argv;
{
  int ch;
  
  while (1)
  {
    if (-1 == (ch = getopt(argc, argv, "")))
      break;
    
    switch (ch)
    {
      
    }
  }
}
**/

int
main (int argc, const char ** argv)
{
  int i = 0;
  daap_song * song;
  http_pool * pool;
  http_lstn * lstn;
  mdns_t * mdns;

  /**
  hash_t * hash;
  hash_key_t * key = NULL;
  
  hash = hash_Create();
  hash_Init(hash);

  hash_Put(hash, "Hello", 5, "1", 1);
  hash_Put(hash, "Hi",    2, "2", 2);
  hash_Put(hash, "How",   3, "3", 3);
  hash_Put(hash, "Are",   3, "4", 4);
  hash_Put(hash, "You",   3, "5", 5);
  hash_Put(hash, "Fine",  4, "6", 6);
  hash_Put(hash, "Thank", 5, "7", 7);
  hash_Put(hash, "You",   3, "8", 8);

  key = hash_Get(hash, "Hello", 5);
  assert(key != NULL);
  fprintf(stderr, "main.. hash_Get('Hello') = '%s' = '%s'\n", key->name, key->data);

  key = hash_Get(hash, "You", 3);
  assert(key != NULL);
  fprintf(stderr, "main.. hash_Get('You') = '%s' = '%s'\n", key->name, key->data);
  
  exit(0);
  **/
  
  /**
  fprintf(stderr, "hash_Hash.. Hello = %lu\n", hash_Hash(hash, "Hello"));
  fprintf(stderr, "hash_Hash.. Hello = %lu\n", hash_Hash(hash, "Hello"));
  fprintf(stderr, "hash_Hash.. abcde = %lu\n", hash_Hash(hash, "abcde"));
  fprintf(stderr, "hash_Hash.. /Users/Shared/iTunes/ = %lu\n", hash_Hash(hash, "/Users/Shared/iTunes/"));
  **/

  /**
  {
    hash_key_t * key = hash_key_Create();
    hash_key_Init(key, "a", 1);
    
    fprintf(stderr, "main.. hash_key_CompareToValue() = %d\n", hash_key_CompareToValue(key, "b", 1));
    fprintf(stderr, "main.. hash_key_CompareToValue() = %d\n", hash_key_CompareToValue(key, "a", 1));
  }
  **/
  
  
  //
  // initialize the persistent song database
  //
  fprintf(stderr, "main.. calling daap_database_Init()\n");
  daap_database_Init(&kDefaultDatabase);

  //
  // load the persistent song database
  //
  if (0 >= (i = daap_database_Import(&kDefaultDatabase))) {
    fprintf(stderr, "main.. calling daap_database_LoadDir(\"./files\")\n");
    daap_database_LoadDir(&kDefaultDatabase, "./files");
    daap_database_Export(&kDefaultDatabase);
  }
  else
    daap_database_Update(&kDefaultDatabase, "./files");

  //
  // initialize the daap song database structure
  //
  fprintf(stderr, "main.. calling adbs_Init()\n");
  adbs_Init(&kDefaultAdbs);

  //
  // initialize the daap song database structure thingy
  //
  fprintf(stderr, "main.. calling apso_Init()\n");
  apso_Init(&kDefaultApso);

  //
  // copy each song from the database into the persistent
  // playlist and song database thingy. mind you, this doesn't
  // actually copy the MP3 ... just various bits of data from
  // the daap_song structure for each song.
  //
  for (i = 0; i < kDefaultDatabase.max; i++) {
    if (NULL == (song = kDefaultDatabase.songs[i]))
      continue;
    
    daap_list_Append((__daap_list *)&kDefaultAdbs.mlcl, (daap_base *)song);
    apso_AddSong(&kDefaultApso, song);
  }

  
  
  signal(SIGPIPE,  main_HandleSigPipe);     // SIGPIPE is a sign of bad things
  signal(SIGHUP,   main_HandleSigHup);      // SIGHUP  has to be sent by kill -HUP <pid>
  signal(SIGINT,   main_HandleSigInt);      // SIGINT  is what you get for a Ctrl-C
  signal(SIGQUIT,  main_HandleSigQuit);     // SIGQUIT is what you get for a Ctrl-
  signal(SIGUSR1,  main_HandleSigUsr1);     // SIGUSR1 has to be sent by kill -USR1 <pid>
  signal(SIGUSR2,  main_HandleSigUsr2);     // SIGUSR2 has to be sent by kill -USR2 <pid>
  
  

  //
  // an http_pool is a collection of http_serf's, which are the worker threads,
  // much like they were the worker class. they do all of the grunt work. the
  //
  fprintf(stderr, "main.. calling http_pool_Create()\n");
  pool = http_pool_Create();
  assert(pool != NULL);
  
  fprintf(stderr, "main.. calling http_pool_Init()\n");
  http_pool_Init(pool);
  
  
  
  //
  // create some number of threads to deal with http requests. threads ARE NOT
  // used on a per-connection basis, like some asinine-ly designed threading
  // models by people who haven't a clue what they are doing. instead, a pool
  // of threads exitsts to do work as work is required - otherwise, there could
  // potentially be 100 connections but every connection remaining utterly 
  // idle ... and 100 threads would also be sitting there idle.
  //
  // TODO: update this to watch how many threads are idle, and dynamically
  // increase or decrease the number of threads based on server load.
  //
  for (i = 0; i < 5; i++) {
    http_serf * serf;
    
    fprintf(stderr, "main.. calling http_serf_Create()\n");
    serf = http_serf_Create();
    assert(serf != NULL);
    
    fprintf(stderr, "main.. calling http_serf_Init()\n");
    http_serf_Init(serf, pool);
    
    fprintf(stderr, "main.. calling http_serf_Start()\n");
    http_serf_Start(serf);
    
    usleep(10 * 5000); // one twentieth of a second
  }
  
  
  
  

  //
  // this is the dude that listens on the port (5090, presently) for
  // incoming socket connections, and dumps them off in a pool to wait
  // for them to send us a request, and be handled by a http_serf thread.
  //
  fprintf(stderr, "main.. calling http_lstn_Create()\n");
  lstn = http_lstn_Create();
  assert(lstn != NULL);
  
  fprintf(stderr, "main.. calling http_lstn_Init()\n");
  http_lstn_Init(lstn);
  
  fprintf(stderr, "main.. calling http_lstn_SetPort()\n");
  http_lstn_SetPort(lstn, 5090);
  
  fprintf(stderr, "main.. calling http_lstn_SetPool()\n");
  http_lstn_SetPool(lstn, pool);
  
  fprintf(stderr, "main.. calling http_lstn_Start()\n");
  http_lstn_Start(lstn);

  
  
  

  //
  // this is a little wrapper around Apple's mDNS Rendezvous implementation.
  // I guess I'll probably get around to replacing it with my own implementation
  // at some point, so that I don't have to deal with their licensing restrictions.
  // until then, this is it.
  //
  fprintf(stderr, "main.. calling mdns_Create()\n");
  mdns = mdns_Create();
  assert(mdns != NULL);
  
  fprintf(stderr, "main.. calling mdns_Init()\n");
  mdns_Init(mdns);
  
  fprintf(stderr, "main.. calling mdns_Start()\n");
  mdns_Start(mdns);

  

  fprintf(stderr, "main.. block indefinitely\n");
  
  while (__hault == 0)
  {
    sleep(1);
  }

  
  
  //
  // tell everyone on the network that we are shutting down, so that our
  // service becomes unavailable immediately, instead of waiting for some
  // timeout to occur on each remote device.
  //
  mdns_Destroy(mdns);

  
  
  // you'll have to figure this one out on your own.
  return 0;
}
